# coding: utf-8
# © 2016-2020 Resurface Labs Inc.

import json
import re
from typing import List, Optional, Pattern

from usagelogger.base_logger import BaseLogger
from usagelogger.http_rules import HttpRules


class HttpLogger(BaseLogger):
    """Usage logger for HTTP/HTTPS protocol."""

    # Agent string identifying this logger.
    AGENT: str = 'http_logger.py'

    def __init__(self, enabled: Optional[bool] = True,
                 queue: Optional[List[str]] = None,
                 url: Optional[str] = None,
                 skip_compression: Optional[bool] = False,
                 skip_submission: Optional[bool] = False,
                 rules: Optional[str] = None) -> None:

        super().__init__(self.AGENT, enabled=enabled, queue=queue, url=url,
                         skip_compression=skip_compression, skip_submission=skip_submission)

        # parse specified rules
        self._rules = HttpRules(rules)

        # apply configuration rules
        self.skip_compression = self._rules.skip_compression
        self.skip_submission = self._rules.skip_submission
        if self._enabled and url is not None and url.startswith('http:') and not self._rules.allow_http_url:
            self._enableable = False
            self._enabled = False

    @property
    def rules(self) -> HttpRules:
        return self._rules

    def submit_if_passing(self, details: List[List[str]]) -> None:
        # apply active rules
        details = self._rules.apply(details)
        if details is None: return

        # finalize message
        details.append(['agent', self.AGENT])
        details.append(['host', self.host])
        details.append(['version', self.version])

        # let's do this thing
        self.submit(json.dumps(details, separators=(',', ':')))

    @classmethod
    def is_string_content_type(cls, s: str) -> bool:
        return s is not None and cls.__STRING_TYPES_REGEX.match(s) is not None

    __STRING_TYPES: str = (r'^(text/(html|plain|xml))|'
                           r'(application/(json|soap|xml|x-www-form-urlencoded))')
    __STRING_TYPES_REGEX: Pattern = re.compile(__STRING_TYPES, flags=re.IGNORECASE)



# coding: utf-8
# © 2016-2020 Resurface Labs Inc.

from django.conf import settings

from usagelogger import HttpLogger, HttpMessage


def __read_settings__(key):
    try:
        return settings.USAGELOGGER[key]
    except Exception:
        return None


class HttpLoggerForDjango:
    def __init__(self, get_response):
        self.get_response = get_response
        self.logger = HttpLogger(url=__read_settings__('url'), rules=__read_settings__('rules'))

    def __call__(self, request):
        response = self.get_response(request)
        status = response.status_code
        if (status < 300 or status == 302) and HttpLogger.is_string_content_type(response['Content-Type']):
            HttpMessage.send(self.logger, request=request, response=response)  # todo add timing details
        return response



# coding: utf-8
# © 2016-2020 Resurface Labs Inc.

from typing import Dict, Optional


class HttpResponseImpl(object):

    def __init__(self, status: Optional[int] = None,
                 headers: Optional[Dict[str, str]] = None,
                 body: Optional[str] = None) -> None:
        self.status = status
        self.headers = {} if headers is None else headers
        self.body = body



# coding: utf-8
# © 2016-2020 Resurface Labs Inc.

import http.client
import os
import socket
import threading
from typing import Dict, List, Optional
from urllib.parse import urlsplit

import usagelogger  # just to read version
from usagelogger.usage_loggers import UsageLoggers


class BaseLogger(object):
    """Basic usage logger to embed or extend."""

    def __init__(self, agent: str,
                 enabled: Optional[bool] = True,
                 queue: Optional[List[str]] = None,
                 url: Optional[str] = None,
                 skip_compression: Optional[bool] = False,
                 skip_submission: Optional[bool] = False) -> None:

        self.agent = agent
        self.host = self.host_lookup()
        self.skip_compression = skip_compression
        self.skip_submission = skip_submission
        self.version = self.version_lookup()

        # read provided options
        if url is None: url = UsageLoggers.url_by_default()

        # set options in priority order
        self._enabled = enabled
        self._queue = queue if isinstance(queue, list) else None
        if self._queue is not None:
            self._url = None
        elif url is not None and isinstance(url, str):
            try:
                if urlsplit(url).scheme in {'http', 'https'}:
                    self._url_scheme: str = urlsplit(url).scheme
                    self._url = url
                else:
                    raise TypeError('incorrect URL scheme')
            except TypeError:
                self._enabled = False
                self._url = None
        else:
            self._enabled = False
            self._url = None

        # finalize internal properties
        self._enableable = self.queue is not None or self.url is not None
        self._submit_failures = 0
        self._submit_failures_lock = threading.Lock()
        self._submit_successes = 0
        self._submit_successes_lock = threading.Lock()

    def disable(self):
        self._enabled = False
        return self

    def enable(self):
        if self.enableable:
            self._enabled = True
        return self

    @property
    def enableable(self) -> bool:
        return self._enableable

    @property
    def enabled(self) -> bool:
        return self._enabled and UsageLoggers.is_enabled()

    @property
    def queue(self) -> List[str]:
        return self._queue

    def submit(self, msg: Optional[str]) -> None:
        """Submits JSON message to intended destination."""

        if msg is None or self.skip_submission is True or self.enabled is False:
            pass
        elif self._queue is not None:
            self._queue.append(msg)
            with self._submit_successes_lock:
                self._submit_successes += 1
        else:
            try:
                # todo implement compression (Clubhouse #49)
                url_parser = urlsplit(self.url)
                hostname = url_parser.hostname  # todo cache this (Clubhouse #49)
                url_path = url_parser.path + url_parser.query  # todo cache this (Clubhouse #49)

                headers: Dict[str, str] = {'Content-Type': 'application/json'}

                if self._url_scheme == "http":
                    conn = http.client.HTTPConnection(hostname)
                else:
                    conn = http.client.HTTPSConnection(hostname)

                conn.request("POST", url_path, msg, headers)
                response = conn.getresponse()
                conn.close()  # todo keep connection alive? (Clubhouse #49)

                if response.status == 204:
                    with self._submit_successes_lock:
                        self._submit_successes += 1
                else:
                    with self._submit_failures_lock:
                        self._submit_failures += 1

            # http errors
            except (http.client.HTTPException, IOError, OSError):
                with self._submit_failures_lock:
                    self._submit_failures += 1
            # JSON errors
            except (OverflowError, TypeError, ValueError):
                with self._submit_failures_lock:
                    self._submit_failures += 1

    @property
    def submit_failures(self) -> int:
        return self._submit_failures

    @property
    def submit_successes(self) -> int:
        return self._submit_successes

    @property
    def url(self) -> str:
        return self._url

    @staticmethod
    def host_lookup() -> str:
        dyno = os.getenv('DYNO')
        if dyno is not None: return dyno
        try:
            return socket.gethostname()
        except:
            return 'unknown'

    @staticmethod
    def version_lookup() -> str:
        return usagelogger.__version__



# coding: utf-8
# © 2016-2020 Resurface Labs Inc.
from typing import Optional, Pattern, Union


class HttpRule(object):

    def __init__(self, verb: str,
                 scope: Optional[Pattern] = None,
                 param1: Optional[Union[Pattern, str, int]] = None,
                 param2: Optional[Union[Pattern, str]] = None) -> None:
        self._verb = verb
        self._scope = scope
        self._param1 = param1
        self._param2 = param2

    @property
    def verb(self):
        return self._verb

    @property
    def scope(self):
        return self._scope

    @property
    def param1(self):
        return self._param1

    @property
    def param2(self):
        return self._param2



# coding: utf-8
# © 2016-2020 Resurface Labs Inc.

import random
import re
from typing import List, Optional, Pattern, Sized

from usagelogger.http_rule import HttpRule


class HttpRules(Sized):
    __DEBUG_RULES: str = (r"allow_http_url" "\n"
                          r"copy_session_field /.*/" "\n")

    __STANDARD_RULES: str = ("/request_header:cookie|response_header:set-cookie/ remove\n"
                             r'/(request|response)_body|request_param/ replace '
                             r'/[a-zA-Z0-9.!#$%&’*+\/=?^_`{|}~-]+@[a-zA-Z0-9-]+'
                             r'(?:\.[a-zA-Z0-9-]+)/, /x@y.com/' "\n"
                             r'/request_body|request_param|response_body/ replace '
                             r'/[0-9\.\-\/]{9,}/, /xyxy/' "\n")

    __STRICT_RULES: str = (r'/request_url/ replace /([^\?;]+).*/, !\\1!' "\n"
                           r'/request_body|response_body|request_param:.*|'
                           r'request_header:(?!user-agent).*|response_header:'
                           r'(?!(content-length)|(content-type)).*/ remove' "\n")

    __default_rules: str = __STRICT_RULES

    @classmethod
    def default_rules(cls) -> str:
        return cls.__default_rules

    @classmethod
    def set_default_rules(cls, rules: str) -> None:
        cls.__default_rules: str = re.sub(r'^\s*include default\s*$', '', rules, flags=re.MULTILINE)

    @classmethod
    def debug_rules(cls) -> str:
        return cls.__DEBUG_RULES

    @classmethod
    def standard_rules(cls) -> str:
        return cls.__STANDARD_RULES

    @classmethod
    def strict_rules(cls) -> str:
        return cls.__STRICT_RULES

    @classmethod
    def parse_rule(cls, rule: str) -> Optional[HttpRule]:
        """Parses rule from single line."""
        if rule is None or HttpRules.__REGEX_BLANK_OR_COMMENT.match(rule): return None
        m = HttpRules.__REGEX_ALLOW_HTTP_URL.match(rule)
        if m: return HttpRule('allow_http_url')
        m = HttpRules.__REGEX_COPY_SESSION_FIELD.match(rule)
        if m: return HttpRule('copy_session_field', None, HttpRules.parse_regex(rule, m.group(1)))
        m = HttpRules.__REGEX_REMOVE.match(rule)
        if m: return HttpRule('remove', HttpRules.parse_regex(rule, m.group(1)))
        m = HttpRules.__REGEX_REMOVE_IF.match(rule)
        if m: return HttpRule('remove_if', HttpRules.parse_regex(rule, m.group(1)), HttpRules.parse_regex(rule, m.group(2)))
        m = HttpRules.__REGEX_REMOVE_IF_FOUND.match(rule)
        if m: return HttpRule('remove_if_found', HttpRules.parse_regex(rule, m.group(1)),
                              HttpRules.parse_regex_find(rule, m.group(2)))
        m = HttpRules.__REGEX_REMOVE_UNLESS.match(rule)
        if m: return HttpRule('remove_unless', HttpRules.parse_regex(rule, m.group(1)), HttpRules.parse_regex(rule, m.group(2)))
        m = HttpRules.__REGEX_REMOVE_UNLESS_FOUND.match(rule)
        if m: return HttpRule('remove_unless_found', HttpRules.parse_regex(rule, m.group(1)),
                              HttpRules.parse_regex_find(rule, m.group(2)))
        m = HttpRules.__REGEX_REPLACE.match(rule)
        if m: return HttpRule('replace', HttpRules.parse_regex(rule, m.group(1)), HttpRules.parse_regex_find(rule, m.group(2)),
                              HttpRules.parse_string(rule, m.group(3)))
        m = HttpRules.__REGEX_SAMPLE.match(rule)
        if m:
            m1 = int(m.group(1))
            if m1 < 1 or m1 > 99: raise SyntaxError('Invalid sample percent: ' + m.group(1))
            return HttpRule('sample', None, m1)
        m = HttpRules.__REGEX_SKIP_COMPRESSION.match(rule)
        if m: return HttpRule('skip_compression')
        m = HttpRules.__REGEX_SKIP_SUBMISSION.match(rule)
        if m: return HttpRule('skip_submission')
        m = HttpRules.__REGEX_STOP.match(rule)
        if m: return HttpRule('stop', HttpRules.parse_regex(rule, m.group(1)))
        m = HttpRules.__REGEX_STOP_IF.match(rule)
        if m: return HttpRule('stop_if', HttpRules.parse_regex(rule, m.group(1)), HttpRules.parse_regex(rule, m.group(2)))
        m = HttpRules.__REGEX_STOP_IF_FOUND.match(rule)
        if m: return HttpRule('stop_if_found', HttpRules.parse_regex(rule, m.group(1)),
                              HttpRules.parse_regex_find(rule, m.group(2)))
        m = HttpRules.__REGEX_STOP_UNLESS.match(rule)
        if m: return HttpRule('stop_unless', HttpRules.parse_regex(rule, m.group(1)), HttpRules.parse_regex(rule, m.group(2)))
        m = HttpRules.__REGEX_STOP_UNLESS_FOUND.match(rule)
        if m: return HttpRule('stop_unless_found', HttpRules.parse_regex(rule, m.group(1)),
                              HttpRules.parse_regex_find(rule, m.group(2)))
        raise SyntaxError('Invalid rule: ' + rule)

    @classmethod
    def parse_regex(cls, rule: str, regex: str) -> Pattern:
        """Parses regex for matching."""
        s: str = HttpRules.parse_string(rule, regex)
        if not s.startswith('^'): s = '^' + s
        if not s.endswith('$'): s = s + '$'
        try:
            return re.compile(s)
        except:
            raise SyntaxError(f'Invalid regex ({regex}) in rule: {rule}')

    @classmethod
    def parse_regex_find(cls, rule: str, regex: str) -> Pattern:
        """Parses regex for finding."""
        try:
            return re.compile(HttpRules.parse_string(rule, regex))
        except:
            raise SyntaxError(f'Invalid regex ({regex}) in rule: {rule}')

    @classmethod
    def parse_string(cls, rule: str, expr: str) -> str:
        """Parses delimited string expression."""
        for sep in ['~', '!', '%', '|', '/']:
            p: Pattern = re.compile(r'^[{s}](.*)[{s}]$'.format(s=sep))
            m = p.match(expr)
            if m:
                m1: str = m.group(1)
                m1p: Pattern = re.compile(r'^[{s}].*|.*[^\\\\][{s}].*'.format(s=sep))
                if m1p.match(m1): raise SyntaxError(f'Unescaped separator ({sep}) in rule: {rule}')
                return sep.join(m1.split('\\' + sep))
        raise SyntaxError(f'Invalid expression ({expr}) in rule: {rule}')

    def __init__(self, rules: str) -> None:
        if rules is None: rules = HttpRules.default_rules()

        # todo load rules from external files

        # force default rules if necessary
        rules = re.sub(r'^\s*include default\s*$', str(HttpRules.default_rules()), rules, flags=re.MULTILINE)
        if len(rules.strip()) == 0: rules = HttpRules.default_rules()

        # expand rule includes
        rules = re.sub(r'(?m)^\s*include debug\s*$', HttpRules.debug_rules(), rules)
        rules = re.sub(r'(?m)^\s*include standard\s*$', HttpRules.standard_rules(), rules)
        rules = re.sub(r'(?m)^\s*include strict\s*$', HttpRules.strict_rules(), rules)
        self._text = rules

        # parse all rules
        prs: List = []
        for rule in rules.split("\n"):
            parsed: HttpRule = HttpRules.parse_rule(rule)
            if parsed is not None: prs.append(parsed)
        self._length = len(prs)

        # break out rules by verb
        self._allow_http_url: bool = len([r for r in prs if 'allow_http_url' == r.verb]) > 0
        self._copy_session_field: List[HttpRule] = [r for r in prs if 'copy_session_field' == r.verb]
        self._remove: List[HttpRule] = [r for r in prs if 'remove' == r.verb]
        self._remove_if: List[HttpRule] = [r for r in prs if 'remove_if' == r.verb]
        self._remove_if_found: List[HttpRule] = [r for r in prs if 'remove_if_found' == r.verb]
        self._remove_unless: List[HttpRule] = [r for r in prs if 'remove_unless' == r.verb]
        self._remove_unless_found: List[HttpRule] = [r for r in prs if 'remove_unless_found' == r.verb]
        self._replace: List[HttpRule] = [r for r in prs if 'replace' == r.verb]
        self._sample: List[HttpRule] = [r for r in prs if 'sample' == r.verb]
        self._skip_compression: bool = len([r for r in prs if r.verb == 'skip_compression']) > 0
        self._skip_submission: bool = len([r for r in prs if r.verb == 'skip_submission']) > 0
        self._stop: List[HttpRule] = [r for r in prs if 'stop' == r.verb]
        self._stop_if: List[HttpRule] = [r for r in prs if 'stop_if' == r.verb]
        self._stop_if_found: List[HttpRule] = [r for r in prs if 'stop_if_found' == r.verb]
        self._stop_unless: List[HttpRule] = [r for r in prs if 'stop_unless' == r.verb]
        self._stop_unless_found: List[HttpRule] = [r for r in prs if 'stop_unless_found' == r.verb]

        # validate rules
        if len(self._sample) > 1: raise SyntaxError('Multiple sample rules')

    def __len__(self) -> int:
        return self._length

    @property
    def allow_http_url(self) -> bool:
        return self._allow_http_url

    @property
    def copy_session_field(self) -> List[HttpRule]:
        return self._copy_session_field

    @property
    def remove(self) -> List[HttpRule]:
        return self._remove

    @property
    def remove_if(self) -> List[HttpRule]:
        return self._remove_if

    @property
    def remove_if_found(self) -> List[HttpRule]:
        return self._remove_if_found

    @property
    def remove_unless(self) -> List[HttpRule]:
        return self._remove_unless

    @property
    def remove_unless_found(self) -> List[HttpRule]:
        return self._remove_unless_found

    @property
    def replace(self) -> List[HttpRule]:
        return self._replace

    @property
    def sample(self) -> List[HttpRule]:
        return self._sample

    @property
    def skip_compression(self) -> bool:
        return self._skip_compression

    @property
    def skip_submission(self) -> bool:
        return self._skip_submission

    @property
    def stop(self) -> List[HttpRule]:
        return self._stop

    @property
    def stop_if(self) -> List[HttpRule]:
        return self._stop_if

    @property
    def stop_if_found(self) -> List[HttpRule]:
        return self._stop_if_found

    @property
    def stop_unless(self) -> List[HttpRule]:
        return self._stop_unless

    @property
    def stop_unless_found(self) -> List[HttpRule]:
        return self._stop_unless_found

    @property
    def text(self) -> str:
        return self._text

    def apply(self, details: List[List[str]]) -> Optional[List[List[str]]]:
        # stop rules come first
        for r in self._stop:
            for d in details:
                if r.scope.match(d[0]): return None
        for r in self._stop_if_found:
            for d in details:
                if r.scope.match(d[0]) and r.param1.search(d[1]): return None
        for r in self._stop_if:
            for d in details:
                if r.scope.match(d[0]) and r.param1.match(d[1]): return None
        passed = 0
        for r in self._stop_unless_found:
            for d in details:
                if r.scope.match(d[0]) and r.param1.search(d[1]): passed += 1
        if passed != len(self._stop_unless_found): return None
        passed = 0
        for r in self._stop_unless:
            for d in details:
                if r.scope.match(d[0]) and r.param1.match(d[1]): passed += 1
        if passed != len(self._stop_unless): return None

        # do sampling if configured
        if len(self._sample) == 1 and random.randrange(100) >= int(self._sample[0].param1): return None

        # winnow sensitive details based on remove rules if configured
        for r in self._remove:
            for d in details:
                if r.scope.match(d[0]): d[1] = ''
        for r in self._remove_unless_found:
            for d in details:
                if r.scope.match(d[0]) and not r.param1.search(d[1]): d[1] = ''
        for r in self._remove_if_found:
            for d in details:
                if r.scope.match(d[0]) and r.param1.search(d[1]): d[1] = ''
        for r in self._remove_unless:
            for d in details:
                if r.scope.match(d[0]) and not r.param1.match(d[1]): d[1] = ''
        for r in self._remove_if:
            for d in details:
                if r.scope.match(d[0]) and r.param1.match(d[1]): d[1] = ''

        # remove any details with empty values
        details = [x for x in details if x[1] != '']
        if len(details) == 0: return None

        # mask sensitive details based on replace rules if configured
        for r in self._replace:
            for d in details:
                if r.scope.match(d[0]): d[1] = re.sub(r.param1, r.param2, d[1])

        # remove any details with empty values
        details = [x for x in details if x[1] != '']
        return None if len(details) == 0 else details

    __REGEX_ALLOW_HTTP_URL: Pattern = re.compile(
        r'^\s*allow_http_url\s*(#.*)?$')
    __REGEX_BLANK_OR_COMMENT: Pattern = re.compile(
        r'^\s*([#].*)*$')
    __REGEX_COPY_SESSION_FIELD: Pattern = re.compile(
        r'^\s*copy_session_field\s+([~!%|/].+[~!%|/])\s*(#.*)?$')
    __REGEX_REMOVE: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*remove\s*(#.*)?$')
    __REGEX_REMOVE_IF: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*remove_if\s+([~!%|/].+[~!%|/])\s*(#.*)?$')
    __REGEX_REMOVE_IF_FOUND: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*'
        r'remove_if_found\s+([~!%|/].+[~!%|/])\s*(#.*)?$')
    __REGEX_REMOVE_UNLESS: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*'
        r'remove_unless\s+([~!%|/].+[~!%|/])\s*(#.*)?$')
    __REGEX_REMOVE_UNLESS_FOUND: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*'
        r'remove_unless_found\s+([~!%|/].+[~!%|/])\s*(#.*)?$')
    __REGEX_REPLACE: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*'
        r'replace[\s]+([~!%|/].+[~!%|/]),[\s]+([~!%|/].*[~!%|/])\s*(#.*)?$')
    __REGEX_SAMPLE: Pattern = re.compile(
        r'^\s*sample\s+(\d+)\s*(#.*)?$')
    __REGEX_SKIP_COMPRESSION: Pattern = re.compile(
        r'^\s*skip_compression\s*(#.*)?$')
    __REGEX_SKIP_SUBMISSION: Pattern = re.compile(
        r'^\s*skip_submission\s*(#.*)?$')
    __REGEX_STOP: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*stop\s*(#.*)?$')
    __REGEX_STOP_IF: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*stop_if\s+([~!%|/].+[~!%|/])\s*(#.*)?$')
    __REGEX_STOP_IF_FOUND: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*'
        r'stop_if_found\s+([~!%|/].+[~!%|/])\s*(#.*)?$')
    __REGEX_STOP_UNLESS: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*stop_unless\s+([~!%|/].+[~!%|/])\s*(#.*)?$')
    __REGEX_STOP_UNLESS_FOUND: Pattern = re.compile(
        r'^\s*([~!%|/].+[~!%|/])\s*'
        r'stop_unless_found\s+([~!%|/].+[~!%|/])\s*(#.*)?$')



# coding: utf-8
# © 2016-2020 Resurface Labs Inc.

import os


class UsageLoggers(object):
    __BRICKED: bool = os.getenv('USAGE_LOGGERS_DISABLE') == 'true'

    __disabled: bool = __BRICKED

    @classmethod
    def disable(cls) -> None:
        cls.__disabled = True

    @classmethod
    def enable(cls) -> None:
        if cls.__BRICKED is False: cls.__disabled = False

    @classmethod
    def is_enabled(cls) -> bool:
        return not cls.__disabled

    @staticmethod
    def url_by_default() -> str:
        return os.getenv('USAGE_LOGGERS_URL')



from .usage_loggers import UsageLoggers
from .http_request_impl import HttpRequestImpl
from .http_response_impl import HttpResponseImpl
from .http_rules import HttpRules
from .base_logger import BaseLogger
from .http_logger import HttpLogger
from .http_message import HttpMessage

# don't import django logger here! (optional, not required)

__version__ = '2.0.0'



from setuptools import setup

setup(
    name='usagelogger',
    version='2.0.0',
    description='Logging usage of Python-based services, with user privacy by design.',
    long_description="""Logging usage of Python-based services, with user privacy by design.

Visit <a href="https://resurface.io">resurface.io</a> for general information on usage logging.

## Middleware Integrations

Loggers can be added to <a href="https://github.com/resurfaceio/logger-python#logging-from-django">Django</a> 
applications with only minor configuration changes.

## Logging With API

Loggers can be directly integrated into your application via our logging
<a href="https://github.com/resurfaceio/logger-python/blob/master/API.md">API</a>.
This requires more effort than middleware integrations, but also offers the greatest flexibility and control.

## Protecting User Privacy

Loggers always have an active set of <a href="https://resurface.io/rules.html">rules</a> that control what data is logged
and how sensitive data is masked. By default, loggers apply a strict set of predefined rules, but logging rules are easily 
customized to meet the needs of any application.

---
&copy; 2016-2020 <a href="https://resurface.io">Resurface Labs Inc.</a>
""",
    long_description_content_type='text/markdown',
    url='https://resurface.io',
    author='Resurface Labs',
    author_email='admin@resurface.io',
    license="Apache License 2.0",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: System Administrators',
        'Topic :: Internet :: Log Analysis',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    keywords='logging resurface',
    packages=['usagelogger'],
    python_requires='>=3.6, <4',
    install_requires=[],
    tests_require=['pytest'],
    project_urls={
        'Bug Reports': 'https://github.com/resurfaceio/logger-python/issues',
        'Source': 'https://github.com/resurfaceio/logger-python'
    }
)



# coding: utf-8
# © 2016-2020 Resurface Labs Inc.

from time import time
from typing import List, Optional

from usagelogger import HttpLogger


class HttpMessage(object):

    @classmethod
    def send(cls, logger: HttpLogger, request, response, response_body: Optional[str] = None,
             request_body: Optional[str] = None, now=None, interval=None) -> None:  # todo missing type hints

        if not logger.enabled: return

        # copy details from request & resonse
        message = cls.build(request, response, response_body, request_body)

        # todo copy details from active session

        # add timing details
        message.append(['now', str(now) if now is not None else str(round(time() * 1000))])
        if interval is not None: message.append(['interval', interval])

        logger.submit_if_passing(message)

    @classmethod
    def build(cls, request, response, response_body: Optional[str] = None,
              request_body: Optional[str] = None) -> Optional[List[List[str]]]:

        message: Optional[List[List[str]]] = None

        if request.__class__.__name__ == "WSGIRequest":
            message = []
            if request.method: message.append(['request_method', request.method])
            url = request.build_absolute_uri()
            if url: message.append(['request_url', url])
            if response.status_code: message.append(['response_code', str(response.status_code)])
            for k, v in request.headers.items(): message.append([f"request_header:{k}".lower(), v])
            if request.method == "GET":
                for k, v in request.GET.items(): message.append([f"request_param:{k}".lower(), v])
            elif request.method == "POST":
                for k, v in request.POST.items(): message.append([f"request_param:{k}".lower(), v])
            for k, v in response.items(): message.append([f"response_header:{k}".lower(), v])
            message.append(['response_body', response.content.decode('utf8')])

        elif request.__class__.__name__ == "HttpRequestImpl":
            message = []
            if request.method: message.append(['request_method', request.method])
            if request.url: message.append(['request_url', request.url])
            if response.status: message.append(['response_code', str(response.status)])
            for k, v in request.headers.items(): message.append([f"request_header:{k}".lower(), v])
            for k, v in request.params.items(): message.append([f"request_param:{k}".lower(), v])
            for k, v in response.headers.items(): message.append([f"response_header:{k}".lower(), v])
            final_request_body = request_body if (request_body is not None) else request.body
            if final_request_body: message.append(['request_body', final_request_body])
            final_response_body = response_body if (response_body is not None) else response.body
            if final_response_body: message.append(['response_body', final_response_body])

        return message
